@@ -3,15 +3,15 @@
 
 package Net::SNMP::Dispatcher;
 
-# $Id: Dispatcher.pm,v 3.1 2005/10/20 14:17:01 dtown Rel $
+# $Id: Dispatcher.pm,v 4.0 2009/09/09 15:05:33 dtown Rel $
 
-# Object the dispatches SNMP messages and handles the scheduling of events.
+# Object that dispatches SNMP messages and handles the scheduling of events.
 
-# Copyright (c) 2001-2005 David M. Town <dtown@cpan.org>
+# Copyright (c) 2001-2009 David M. Town <dtown@cpan.org>
 # All rights reserved.
 
 # This program is free software; you may redistribute it and/or modify it
-# under the same terms as Perl itself.
+# under the same terms as the Perl 5 programming language system itself.
 
 # ============================================================================
 
@@ -23,7 +23,7 @@ use Net::SNMP::Message qw( TRUE FALSE );
 
 ## Version of the Net::SNMP::Dispatcher module
 
-our $VERSION = v3.0.1;
+our $VERSION = v4.0.0;
 
 ## Package variables
 
@@ -35,24 +35,32 @@ our $MESSAGE_PROCESSING;  # Reference to the Message Processing object
 
 ## Event array indexes
 
-sub _ACTIVE()   { 0 }     # State of the event
-sub _TIME()     { 1 }     # Execution time
-sub _CALLBACK() { 2 }     # Callback reference
-sub _PREVIOUS() { 3 }     # Previous event
-sub _NEXT()     { 4 }     # Next event
+sub _ACTIVE    { 0 }      # State of the event
+sub _TIME      { 1 }      # Execution time
+sub _CALLBACK  { 2 }      # Callback reference
+sub _PREVIOUS  { 3 }      # Previous event
+sub _NEXT      { 4 }      # Next event
 
 BEGIN
 {
-   # Use a higher resolution of time() if the Time::HiRes module is available. 
+   # Use a higher resolution of time() and possibly a monotonically
+   # increasing time value if the Time::HiRes module is available. 
 
-   if (eval('require Time::HiRes')) {
+   if (eval 'require Time::HiRes') {
       Time::HiRes->import('time');
+      no warnings;
+      if (eval 'Time::HiRes::clock_gettime(Time::HiRes::CLOCK_MONOTONIC())') {
+         *time = sub () {
+            Time::HiRes::clock_gettime(Time::HiRes::CLOCK_MONOTONIC());
+         };
+      }
    }
 
    # Validate the creation of the Message Processing object. 
 
-   if (!defined($MESSAGE_PROCESSING = Net::SNMP::MessageProcessing->instance)) {
-      die('FATAL: Failed to create Message Processing instance');
+   if (!defined($MESSAGE_PROCESSING = Net::SNMP::MessageProcessing->instance()))
+   {
+      die 'FATAL: Failed to create Message Processing instance';
    }
 }
 
@@ -60,79 +68,48 @@ BEGIN
 
 sub instance
 {
-   $INSTANCE ||= Net::SNMP::Dispatcher->_new;
+   return $INSTANCE ||= Net::SNMP::Dispatcher->_new();
 }
 
-sub activate
+sub loop
 {
    my ($this) = @_;
 
-   # Return immediately if the Dispatcher is already active.
    return TRUE if ($this->{_active});
 
-   # Indicate that the Dispatcher is active and block  
-   # on select() calls.  
+   $this->{_active} = TRUE;
 
-   $this->{_active}   = TRUE;
-   $this->{_blocking} = TRUE;
-
-   while (defined($this->{_event_queue_h})) { 
-      $this->_event_handle; 
+   # Process while there are events and file descriptor handlers.
+   while (defined $this->{_event_queue_h} || keys %{$this->{_descriptors}}) {
+      $this->_event_handle(undef);
    }
 
-   # Flag the Dispatcher as not active 
-   $this->{_active} = FALSE; 
+   return $this->{_active} = FALSE;
 }
 
 sub one_event
 {
    my ($this) = @_;
 
-   # Return immediately if the Dispatcher is already active.
    return TRUE if ($this->{_active});
 
-   # Indicate that the Dispatcher is active and DO NOT 
-   # block on select() calls.
-   
-   $this->{_active}   = TRUE; 
-   $this->{_blocking} = FALSE;
+   if (defined $this->{_event_queue_h} || keys %{$this->{_descriptors}}) {
+      $this->{_active} = TRUE;
+      $this->_event_handle(0);
+      $this->{_active} = FALSE;
+   }
 
-   $this->_event_handle;
+   return (defined $this->{_event_queue_h} || keys %{$this->{_descriptors}});
+}
 
-   # Flag the Dispatcher as not active
-   $this->{_active} = FALSE;
-
-   defined($this->{_event_queue_h}) ? TRUE : FALSE; 
+sub activate
+{
+   goto &loop;
 }
 
 sub listen
 {
-   my ($this) = @_;
-
-   # Return immediately if the Dispatcher is already active.
-   return TRUE if ($this->{_active});
-
-   # Indicate that the Dispatcher is active and block
-   # on select() calls.
-
-   $this->{_active}   = TRUE;
-   $this->{_blocking} = TRUE;
-
-   while (defined($this->{_event_queue_h}) || keys(%{$this->{_descriptors}})) {
-
-      # Handle queued events
-      events: while (defined($this->{_event_queue_h})) {
-         $this->_event_handle;
-      }
-
-      # Block on select() until there is file descriptor activity.
-      DEBUG_INFO('waiting for activity');
-      $this->_event_select(undef);
-
-   }
-
-   # Flag the Dispatcher as not active
-   $this->{_active} = FALSE;
+   goto &loop;
 }
 
 sub send_pdu
@@ -140,10 +117,10 @@ sub send_pdu
    my ($this, $pdu, $delay) = @_;
 
    # Clear any previous errors
-   $this->_error_clear;
+   $this->_error_clear();
 
-   if ((@_ < 2) || !ref($pdu)) {
-      return $this->_error('Required PDU missing');
+   if ((@_ < 2) || !ref $pdu) {
+      return $this->_error('The required PDU object is missing or invalid');
    }
 
    # If the Dispatcher is active and the delay value is negative,
@@ -151,35 +128,40 @@ sub send_pdu
 
    if ($delay < 0) {
       if ($this->{_active}) {
-         return $this->_send_pdu($pdu, $pdu->timeout, $pdu->retries);
-      } 
+         return $this->_send_pdu($pdu, $pdu->retries());
+      }
       $delay = 0;
    }
 
-   $this->schedule($delay, [\&_send_pdu, $pdu, $pdu->timeout, $pdu->retries]);
+   $this->schedule($delay, [\&_send_pdu, $pdu, $pdu->retries()]);
 
-   TRUE;
+   return TRUE;
 }
 
 sub return_response_pdu
 {
    my ($this, $pdu) = @_;
 
-   $this->send_pdu($pdu, -1);
+   return $this->send_pdu($pdu, -1);
+}
+
+sub msg_handle_alloc
+{
+   return $MESSAGE_PROCESSING->msg_handle_alloc();
 }
 
 sub schedule
 {
    my ($this, $time, $callback) = @_;
 
-   $this->_event_create($time, $this->_callback_create($callback));
+   return $this->_event_create($time, $this->_callback_create($callback));
 }
 
 sub cancel
 {
    my ($this, $event) = @_;
 
-   $this->_event_delete($event);
+   return $this->_event_delete($event);
 }
 
 sub register
@@ -189,37 +171,39 @@ sub register
    # Transport Domain and file descriptor must be valid.
    my $fileno;
 
-   if ((!defined($transport)) || (!defined($fileno = $transport->fileno))) {
-      return $this->_error('Invalid Transport Domain');
+   if (!defined($transport) || !defined($fileno = $transport->fileno())) {
+      return $this->_error('The Transport Domain object is invalid');
    }
 
    # NOTE: The callback must read the data associated with the
    #       file descriptor or the Dispatcher will continuously
    #       call the callback and get stuck in an infinite loop.
 
-   if (!exists($this->{_descriptors}->{$fileno})) {
+   if (!exists $this->{_descriptors}->{$fileno}) {
 
-      DEBUG_INFO('adding descriptor [%d]', $fileno);
+      # Make sure that the "readable" vector is defined.
+      if (!defined $this->{_rin}) {
+         $this->{_rin} = q{};
+      }
 
-      $this->{_rin} = '' unless defined($this->{_rin});
-
-      # Add the file descriptor to the list
+      # Add the file descriptor to the list.
       $this->{_descriptors}->{$fileno} = [
          $this->_callback_create($callback), # Callback
          $transport,                         # Transport Domain object 
          1                                   # Reference count
       ];
 
-      # Add the file descriptor to the "readable" vector
+      # Add the file descriptor to the "readable" vector.
       vec($this->{_rin}, $fileno, 1) = 1;
 
+      DEBUG_INFO('added handler for descriptor [%d]', $fileno);
+
    } else {
-      # Bump up the reference count
+      # Bump up the reference count.
       $this->{_descriptors}->{$fileno}->[2]++;
    }
 
-   # Return the Transport Domain object 
-   $transport;
+   return $transport;
 }
 
 sub deregister
@@ -229,45 +213,46 @@ sub deregister
    # Transport Domain and file descriptor must be valid.
    my $fileno;
 
-   if ((!defined($transport)) || (!defined($fileno = $transport->fileno))) {
-      return $this->_error('Invalid Transport Domain');
+   if (!defined($transport) || !defined($fileno = $transport->fileno())) {
+      return $this->_error('The Transport Domain object is invalid');
    }
 
-   if (exists($this->{_descriptors}->{$fileno})) {
+   if (exists $this->{_descriptors}->{$fileno}) {
 
-      # Check reference count
+      # Check reference count.
       if (--$this->{_descriptors}->{$fileno}->[2] < 1) {
 
-         DEBUG_INFO('removing descriptor [%d]', $fileno);
+         # Remove the file descriptor from the list.
+         delete $this->{_descriptors}->{$fileno};
 
-         # Remove the file descriptor from the list
-         delete($this->{_descriptors}->{$fileno});
-
-         # Remove the file descriptor from the "readable" vector
+         # Remove the file descriptor from the "readable" vector.
          vec($this->{_rin}, $fileno, 1) = 0;
 
          # Undefine the vector if there are no file descriptors,
          # some systems expect this to make select() work properly.
 
-         $this->{_rin} = undef unless keys(%{$this->{_descriptors}});
+         if (!keys %{$this->{_descriptors}}) {
+            $this->{_rin} = undef;
+         }
+
+         DEBUG_INFO('removed handler for descriptor [%d]', $fileno);
       }
 
    } else {
-      return $this->_error('Not registered for this Transport Domain');
+      return $this->_error('The Transport Domain object is not registered');
    }
 
-   # Return the Transport Domain object 
-   $transport;
+   return $transport;
 }
 
 sub error
 {
-   $_[0]->{_error} || '';
+   return $_[0]->{_error} || q{};
 }
 
 sub debug
 {
-   (@_ == 2) ? $DEBUG = ($_[1]) ? TRUE : FALSE : $DEBUG;
+   return (@_ == 2) ? $DEBUG = ($_[1]) ? TRUE : FALSE : $DEBUG;
 }
 
 # [private methods] ----------------------------------------------------------
@@ -279,9 +264,8 @@ sub _new
    # The constructor is private since we only want one 
    # Dispatcher object.
 
-   bless {
+   return bless {
       '_active'        => FALSE,  # State of this Dispatcher object
-      '_blocking'      => TRUE,   # Block on select()
       '_error'         => undef,  # Error message
       '_event_queue_h' => undef,  # Head of the event queue
       '_event_queue_t' => undef,  # Tail of the event queue
@@ -292,80 +276,81 @@ sub _new
 
 sub _send_pdu
 {
-   my ($this, $pdu, $timeout, $retries) = @_;
+   my ($this, $pdu, $retries) = @_;
 
    # Pass the PDU to Message Processing so that it can
    # create the new outgoing message.
 
    my $msg = $MESSAGE_PROCESSING->prepare_outgoing_msg($pdu);
 
-   if (!defined($msg)) {
+   if (!defined $msg) {
       # Inform the command generator about the Message Processing error.
-      $pdu->status_information($MESSAGE_PROCESSING->error);
-      return; 
+      $pdu->status_information($MESSAGE_PROCESSING->error());
+      return;
    }
 
    # Actually send the message.
 
-   if (!defined($msg->send)) {
+   if (!defined $msg->send()) {
 
-      if ($pdu->expect_response) {
-         $MESSAGE_PROCESSING->msg_handle_delete($pdu->msg_id);
+      # Delete the msgHandle.
+      if ($pdu->expect_response()) {
+         $MESSAGE_PROCESSING->msg_handle_delete($msg->msg_id());
       }
 
       # A crude attempt to recover from temporary failures.
       if (($retries-- > 0) && ($!{EAGAIN} || $!{EWOULDBLOCK})) {
          DEBUG_INFO('attempting recovery from temporary failure');
-         $this->schedule($timeout, [\&_send_pdu, $pdu, $timeout, $retries]);
+         $this->schedule($pdu->timeout(), [\&_send_pdu, $pdu, $retries]);
          return FALSE;
       }
 
       # Inform the command generator about the send() error.
-      $pdu->status_information($msg->error);
+      $pdu->status_information($msg->error());
 
       return;
    }
 
    # Schedule the timeout handler if the message expects a response.
 
-   if ($pdu->expect_response) {
-      $this->register($msg->transport, [\&_transport_response_received]);
+   if ($pdu->expect_response()) {
+      $this->register($msg->transport(), [\&_transport_response_received]);
       $msg->timeout_id(
          $this->schedule(
-            $timeout, [\&_transport_timeout, $pdu, $timeout, $retries] 
+            $pdu->timeout(),
+            [\&_transport_timeout, $pdu, $retries, $msg->msg_id()]
          )
-      ); 
+      );
    }
 
-   TRUE;
+   return TRUE;
 }
 
 sub _transport_timeout
 {
-   my ($this, $pdu, $timeout, $retries) = @_;
+   my ($this, $pdu, $retries, $handle) = @_;
 
-   # Stop waiting for responses
-   $this->deregister($pdu->transport);
+   # Stop waiting for responses.
+   $this->deregister($pdu->transport());
+
+   # Delete the msgHandle.
+   $MESSAGE_PROCESSING->msg_handle_delete($handle);
 
    if ($retries-- > 0) {
 
       # Resend a new message.
-      DEBUG_INFO('retries left %d', $retries); 
-      $this->_send_pdu($pdu, $timeout, $retries);
+      DEBUG_INFO('retries left %d', $retries);
+      return $this->_send_pdu($pdu, $retries);
 
    } else {
 
-      # Delete the msgHandle. 
-      $MESSAGE_PROCESSING->msg_handle_delete($pdu->msg_id);
-
       # Inform the command generator about the timeout. 
       $pdu->status_information(
-          "No response from remote host '%s'", $pdu->hostname
-      ); 
-
+          q{No response from remote host "%s"}, $pdu->hostname()
+      );
       return;
 
-   } 
+   }
 }
 
 sub _transport_response_received
@@ -373,47 +358,53 @@ sub _transport_response_received
    my ($this, $transport) = @_;
 
    # Clear any previous errors
-   $this->_error_clear;
+   $this->_error_clear();
 
-   die('FATAL: Invalid Transport Domain') unless ref($transport);
+   if (!ref $transport) {
+      die 'FATAL: The Transport Domain object is invalid';
+   }
 
    # Create a new Message object to receive the response
    my ($msg, $error) = Net::SNMP::Message->new(-transport => $transport);
 
-   if (!defined($msg)) {
-      die sprintf('Failed to create Message object [%s]', $error);
+   if (!defined $msg) {
+      die sprintf 'Failed to create Message object: %s', $error;
    }
 
    # Read the message from the Transport Layer  
-   if (!defined($msg->recv)) {
-      $this->deregister($transport) unless ($transport->connectionless);
-      return $this->_error($msg->error);
+   if (!defined $msg->recv()) {
+      if (!$transport->connectionless()) {
+         $this->deregister($transport);
+      }
+      return $this->_error($msg->error());
    }
 
    # For connection-oriented Transport Domains, it is possible to
    # "recv" an empty buffer if reassembly is required.
 
-   if (!$msg->length) {
+   if (!$msg->length()) {
       DEBUG_INFO('ignoring zero length message');
       return FALSE;
    }
 
    # Hand the message over to Message Processing.
-   if (!defined($MESSAGE_PROCESSING->prepare_data_elements($msg))) {
-      return $this->_error($MESSAGE_PROCESSING->error);  
+   if (!defined $MESSAGE_PROCESSING->prepare_data_elements($msg)) {
+      return $this->_error($MESSAGE_PROCESSING->error());
    }
 
    # Set the error if applicable. 
-   $msg->error($MESSAGE_PROCESSING->error) if ($MESSAGE_PROCESSING->error);
+   if ($MESSAGE_PROCESSING->error()) {
+      $msg->error($MESSAGE_PROCESSING->error());
+   }
 
    # Cancel the timeout.
-   $this->cancel($msg->timeout_id);
+   $this->cancel($msg->timeout_id());
 
    # Stop waiting for responses.
    $this->deregister($transport);
 
    # Notify the command generator to process the response.
-   $msg->process_response_pdu; 
+   return $msg->process_response_pdu();
 }
 
 sub _event_create
@@ -426,7 +417,7 @@ sub _event_create
    # the event needs to be created such that it will get properly
    # initialized when the Dispatcher is started.
 
-   $this->_event_insert(
+   return $this->_event_insert(
       [
          $this->{_active},                          # State of the object
          $this->{_active} ? time() + $time : $time, # Execution time
@@ -434,7 +425,7 @@ sub _event_create
          undef,                                     # Previous event
          undef,                                     # Next event 
       ]
-   ); 
+   );
 }
 
 sub _event_insert
@@ -444,7 +435,7 @@ sub _event_insert
    # If the head of the list is not defined, we _must_ be the only
    # entry in the list, so create a new head and tail reference.
 
-   if (!defined($this->{_event_queue_h})) {
+   if (!defined $this->{_event_queue_h}) {
       DEBUG_INFO('created new head and tail [%s]', $event);
       return $this->{_event_queue_h} = $this->{_event_queue_t} = $event;
    }
@@ -462,8 +453,7 @@ sub _event_insert
 
       # Search backwards from the tail of the list
 
-      for (my $e = $this->{_event_queue_t}; defined($e); $e = $e->[_PREVIOUS])
-      {
+      for (my $e = $this->{_event_queue_t}; defined $e; $e = $e->[_PREVIOUS]) {
          if ($e->[_TIME] <= $event->[_TIME]) {
             $event->[_PREVIOUS] = $e;
             $event->[_NEXT] = $e->[_NEXT];
@@ -486,7 +476,7 @@ sub _event_insert
 
       # Search forward from the head of the list
 
-      for (my $e = $this->{_event_queue_h}; defined($e); $e = $e->[_NEXT]) {
+      for (my $e = $this->{_event_queue_h}; defined $e; $e = $e->[_NEXT]) {
          if ($e->[_TIME] > $event->[_TIME]) {
             $event->[_NEXT] = $e;
             $event->[_PREVIOUS] = $e->[_PREVIOUS];
@@ -495,7 +485,7 @@ sub _event_insert
                $this->{_event_queue_h} = $event;
             } else {
                DEBUG_INFO('inserted [%s] into list', $event);
-               $e->[_PREVIOUS]->[_NEXT] = $event; 
+               $e->[_PREVIOUS]->[_NEXT] = $event;
             }
             return $e->[_PREVIOUS] = $event;
          }
@@ -506,44 +496,46 @@ sub _event_insert
       $this->{_event_queue_t} = $this->{_event_queue_t}->[_NEXT] = $event;
 
    }
+
+   return $event;
 }
 
 sub _event_delete
 {
    my ($this, $event) = @_;
 
-   my $info = '';
+   my $info = q{};
 
    # Update the previous event
-   if (defined($event->[_PREVIOUS])) {
+   if (defined $event->[_PREVIOUS]) {
       $event->[_PREVIOUS]->[_NEXT] = $event->[_NEXT];
    } elsif ($event eq $this->{_event_queue_h}) {
-      if (defined($this->{_event_queue_h} = $event->[_NEXT])) {
-          $info = sprintf(', defined new head [%s]', $event->[_NEXT]);
+      if (defined ($this->{_event_queue_h} = $event->[_NEXT])) {
+          $info = sprintf ', defined new head [%s]', $event->[_NEXT];
       } else {
          DEBUG_INFO('deleted [%s], list is now empty', $event);
-         $this->{_event_queue_t} = undef @{$event}; 
+         $this->{_event_queue_t} = undef @{$event};
          return FALSE; # Indicate queue is empty
       }
    } else {
-      die('FATAL: Attempt to delete invalid Event head');
+      die 'FATAL: Attempted to delete Event object with an invalid head';
    }
 
    # Update the next event
-   if (defined($event->[_NEXT])) {
+   if (defined $event->[_NEXT]) {
       $event->[_NEXT]->[_PREVIOUS] = $event->[_PREVIOUS];
    } elsif ($event eq $this->{_event_queue_t}) {
-      $info .= sprintf(', defined new tail [%s]', $event->[_PREVIOUS]); 
+      $info .= sprintf ', defined new tail [%s]', $event->[_PREVIOUS];
       $this->{_event_queue_t} = $event->[_PREVIOUS];
    } else {
-      die('FATAL: Attempt to delete invalid Event tail');
+      die 'FATAL: Attempted to delete Event object with an invalid tail';
    }
 
    DEBUG_INFO('deleted [%s]%s', $event, $info);
    undef @{$event};
 
    # Indicate queue still has entries
-   TRUE;
+   return TRUE;
 }
 
 sub _event_init
@@ -560,108 +552,86 @@ sub _event_init
 
    # Update the appropriate fields.
    $event->[_ACTIVE]   = $this->{_active};
-   $event->[_TIME]     = $this->{_active} ? time() + $time : $time; 
+   $event->[_TIME]     = $this->{_active} ? time() + $time : $time;
    $event->[_CALLBACK] = $callback;
 
    # Insert the event back into the queue.
-   $this->_event_insert($event); 
+   $this->_event_insert($event);
 
-   TRUE;
+   return TRUE;
 }
 
 sub _event_handle
-{
-   my ($this) = @_;
-
-   # Events are sorted by time, so the event at the head of the list
-   # is the next event that needs to be executed.
-
-   return FALSE unless defined(my $event = $this->{_event_queue_h});
-
-   # Calculate a timeout based on the current time and the lowest 
-   # event time (if the event does not need initialized).
-
-   my $timeout = ($event->[_ACTIVE]) ? ($event->[_TIME] - time()) : 0;
-
-   # If the timeout is less than 0, we are running late.  Adjust the
-   # the timeout to poll the descriptors before acting on the event. 
-   
-   if ($timeout < 0) {
-      DEBUG_INFO('event [%s], skew = %f', $event, -$timeout);
-      $timeout = 0;
-   } else {
-      DEBUG_INFO('event [%s], poll delay = %f', $event, $timeout);
-   }
-
-   # Check the file descriptors for activity.  If there has been any
-   # activity, we must return because the activity could have cancelled 
-   # the event or returned control here before the event is ready to
-   # be acted upon. 
-
-   return TRUE if ($this->_event_select($this->{_blocking} ? $timeout : 0));
-
-   # If we are not blocking and the timeout is non-zero, then it is
-   # not time to act on the event.
-
-   return TRUE if ((!$this->{_blocking}) && ($timeout > 0));
-
-   # If we made it here, we can finally act on the event.  If the event
-   # was inserted with a non-zero delay while the Dispatcher was not
-   # active, the execution time of the event needs to be updated.
-
-   if ((!$event->[_ACTIVE]) && ($event->[_TIME] > 0)) {
-      return $this->_event_init($event);
-   } else {
-      $this->_callback_execute($event->[_CALLBACK]);
-   }
-
-   # Once we reach here, we are done with the event, so remove it
-   # from the head of the queue.
-
-   $this->_event_delete($event);
-}
-
-sub _event_select
 {
    my ($this, $timeout) = @_;
 
+   my $time = time();
+
+   if (defined (my $event = $this->{_event_queue_h})) {
+
+      # If the event was inserted with a non-zero delay while the
+      # Dispatcher was not active, the scheduled time of the event
+      # needs to be updated.
+
+      if (!$event->[_ACTIVE] && $event->[_TIME]) {
+         return $this->_event_init($event);
+      }
+
+      if ($event->[_TIME] <= $time) {
+
+         # If the scheduled time of the event is past, execute it and
+         # set the timeout to zero to poll the descriptors immediately.
+
+         $this->_callback_execute($event->[_CALLBACK]);
+         $this->_event_delete($event);
+         $timeout = 0;
+
+      } elsif (!defined $timeout) {
+
+         # Calculate the timeout for the next event unless one was
+         # specified by the caller.
+
+         $timeout = $event->[_TIME] - $time;
+         DEBUG_INFO('event [%s], timeout = %.04f', $event, $timeout);
+
+      }
+
+   }
+
+   # Check the file descriptors for activity.
+
    my $nfound = select(my $rout = $this->{_rin}, undef, undef, $timeout);
 
-   if ((!defined($nfound)) || ($nfound < 0)) {
+   if (!defined $nfound || $nfound < 0) {
 
       if ($!{EINTR}) { # Recoverable error
-         return TRUE;
+         return FALSE;
       } else {
-         die sprintf('FATAL: select() error [%s]', $!);
+         die sprintf 'FATAL: select() error: %s', $!;
       }
 
    } elsif ($nfound > 0) {
 
       # Find out which file descriptors have data ready for reading.
 
-      if (defined($rout)) {
-         foreach (keys(%{$this->{_descriptors}})) {
-            if (vec($rout, $_, 1)) {
+      if (defined $rout) {
+         for (keys %{$this->{_descriptors}}) {
+            if (vec $rout, $_, 1) {
                DEBUG_INFO('descriptor [%d] ready for read', $_);
                $this->_callback_execute(@{$this->{_descriptors}->{$_}}[0,1]);
             }
          }
       }
 
-      return TRUE;
    }
 
-   # No file descriptor activity. 
-
-   FALSE; 
+   return TRUE;
 }
 
 sub _callback_create
 {
    my ($this, $callback) = @_;
 
-   return unless (@_ == 2);
-
    # Callbacks can be passed in two different ways.  If the callback
    # has options, the callback must be passed as an ARRAY reference
    # with the first element being a CODE reference and the remaining
@@ -669,44 +639,39 @@ sub _callback_create
    # just passed as a CODE reference.
 
    if ((ref($callback) eq 'ARRAY') && (ref($callback->[0]) eq 'CODE')) {
-      $callback;
+      return $callback;
    } elsif (ref($callback) eq 'CODE') {
-      [$callback];
+      return [$callback];
    } else {
-      return;
+      return [];
    }
 }
 
 sub _callback_execute
 {
-#  my ($this, @argv) = @_;
-
-   return unless (@_ > 1) && defined($_[1]);
+   my ($this, @argv) = @_;
 
    # The callback is invoked passing a reference to this object
-   # with the parameters passed by the user next and then any 
-   # parameters that we provide.
+   # with the parameters passed by the user next and then any
+   # parameters that the caller provides.
 
-   my $this = shift(@_);
-   my @argv = @{shift(@_)};
-   my $cb   = shift(@argv);
-   
-   # Protect ourselves from user error. 
-   eval { $cb->($this, @argv, @_); };
+   my ($callback, @user_argv) = @{shift @argv};
 
-   ($@) ? $this->_error($@) : TRUE;
+   # Protect ourselves from user error.
+   eval { $callback->($this, @user_argv, @argv); };
+
+   return ($@) ? $this->_error($@) : TRUE;
 }
 
 sub _error
 {
    my $this = shift;
 
-   if (!defined($this->{_error})) {
+   if (!defined $this->{_error}) {
       $this->{_error} = (@_ > 1) ? sprintf(shift(@_), @_) : $_[0];
-      if ($this->debug) {
-         printf("error: [%d] %s(): %s\n",
-            (caller(0))[2], (caller(1))[3], $this->{_error}
-         );
+      if ($this->debug()) {
+         printf "error: [%d] %s(): %s\n",
+                (caller 0)[2], (caller 1)[3], $this->{_error};
       }
    }
 
@@ -715,21 +680,18 @@ sub _error
 
 sub _error_clear
 {
-   $_[0]->{_error} = undef;
+   return $_[0]->{_error} = undef;
 }
 
 sub DEBUG_INFO
 {
-   return unless $DEBUG;
+   return $DEBUG if (!$DEBUG);
 
-   printf(
-      sprintf('debug: [%d] %s(): ', (caller(0))[2], (caller(1))[3]) . 
+   return printf
+      sprintf('debug: [%d] %s(): ', (caller 0)[2], (caller 1)[3]) .
       ((@_ > 1) ? shift(@_) : '%s') .
       "\n",
-      @_
-   );
-
-   $DEBUG;
+      @_;
 }
 
 # ============================================================================
